unit quaternion;

{$mode objfpc}{$H+}
//{$DEFINE UsecTypes}

interface

uses
  SysUtils,math{$IFDEF UsecTypes},ctypes{$ENDIF};

Const
  QUATERNION_EPS = 1e-4;
{$IFNDEF UsecTypes}
Type
  cDouble = Double;
  cBool = Boolean;
{$ENDIF}


Type
  PQuaternion = ^TQuaternion;
  TQuaternion = Record
    w:cDouble;
    v:Array[0..2] Of cDouble;
  end;
  TDouble3 = Array[0..2] Of cDouble;

Procedure Quaternion_set(w:cDouble;v1:cDouble;v2:cDouble;v3:cDouble;output:PQuaternion);
Procedure Quaternion_setIdentity(q:PQuaternion);
Procedure Quaternion_copy(q:PQuaternion;output:PQuaternion);
Function Quaternion_equal(q1:PQuaternion;q2:PQuaternion):cbool;
Procedure Quaternion_fprint(filename:Pointer;q:PQuaternion);
Procedure Quaternion_fromAxisAngle(axis:TDouble3;angle:cDouble;output:PQuaternion);
Function Quaternion_toAxisAngle(q:PQuaternion;output:TDouble3):cDouble;
Procedure Quaternion_fromEulerZYX(eulerZYX:TDouble3;output:PQuaternion);
Procedure Quaternion_toEulerZYX(q:PQuaternion;output:TDouble3);
Procedure Quaternion_fromXRotation(angle:cDouble;output:PQuaternion);
Procedure Quaternion_fromYRotation(angle:cDouble;output:PQuaternion);
Procedure Quaternion_fromZRotation(angle:cDouble;output:PQuaternion);
Function Quaternion_norm(q:PQuaternion):cDouble;
Procedure Quaternion_normalize(q:PQuaternion;output:PQuaternion);
Procedure Quaternion_conjugate(q:PQuaternion;output:PQuaternion);
Procedure Quaternion_multiply(q1:PQuaternion;q2:PQuaternion;output:PQuaternion);
Procedure Quaternion_rotate(q:PQuaternion;v:TDouble3;output:TDouble3);
Procedure Quaternion_slerp(q1:PQuaternion;q2:PQuaternion;t:cDouble;output:PQuaternion);

implementation

procedure Quaternion_set(w: cDouble; v1: cDouble; v2: cDouble; v3: cDouble;
  output: PQuaternion);
begin
  If Assigned(output) Then
     Begin
       output^.w:=w;
       output^.v[0]:=v1;
       output^.v[1]:=v2;
       output^.v[2]:=v3;
     end;
end;

procedure Quaternion_setIdentity(q: PQuaternion);
begin
  If Assigned(q) Then Quaternion_set(1,0,0,0,q);
end;

procedure Quaternion_copy(q: PQuaternion; output: PQuaternion);
begin
  Quaternion_set(q^.w,q^.v[0],q^.v[1],q^.v[2],output);
end;

function Quaternion_equal(q1: PQuaternion; q2: PQuaternion): cbool;
Var
  equalW:cbool;
  equalV0:cbool;
  equalV1:cbool;
  equalV2:cbool;
begin
  equalW:=abs(q1^.w-q2^.w)<=QUATERNION_EPS;
  equalV0:=abs(q1^.v[0]-q2^.v[0])<=QUATERNION_EPS;
  equalV1:=abs(q1^.v[1]-q2^.v[1])<=QUATERNION_EPS;
  equalV2:=abs(q1^.v[2]-q2^.v[2])<=QUATERNION_EPS;
  Result:=equalW And equalV0 And equalV1 And equalV2;
end;

procedure Quaternion_fprint(filename: Pointer; q: PQuaternion);
begin
  //Change it to write to file
  Writeln(Format('(%.3f, %.3f, %.3f, %.3f)',[q^.w,q^.v[0],q^.v[1],q^.v[2]]));
end;

procedure Quaternion_fromAxisAngle(axis: TDouble3; angle: cDouble;
  output: PQuaternion);
Var
  c:cDouble;
begin
  If Assigned(output) Then
     Begin
       output^.w:=cos(angle/2.0);
       c:=sin(angle/2.0);
       output^.v[0]:=c*axis[0];
       output^.v[1]:=c*axis[1];
       output^.v[2]:=c*axis[2];
     end;
end;

function Quaternion_toAxisAngle(q: PQuaternion; output: TDouble3): cDouble;
Var
  angle:cDouble;      //Local var neccessary?
  divider:cDouble;
begin
  If Assigned(@output[0]) Then
     Begin
       angle:=2.0*arccos(q^.w);
       divider:=sqrt(1.0-q^.w*q^.w);
       If divider<>0.0 Then
          Begin
            output[0]:=q^.v[0]/divider;
            output[1]:=q^.v[1]/divider;
            output[2]:=q^.v[2]/divider;
          end
       Else
         Begin
           output[0]:=1;
           output[1]:=0;
           output[2]:=0;
         end;
     end;
  Result:=angle;
end;

procedure Quaternion_fromEulerZYX(eulerZYX: TDouble3; output: PQuaternion);
Var
  cy:cDouble;
  sy:cDouble;
  cr:cDouble;
  sr:cDouble;
  cp:cDouble;
  sp:cDouble;
begin
  If Assigned(output) Then
     Begin
       cy:=cos(eulerZYX[2]*0.5);
       sy:=sin(eulerZYX[2]*0.5);
       cr:=cos(eulerZYX[0]*0.5);
       sr:=sin(eulerZYX[0]*0.5);
       cp:=cos(eulerZYX[1]*0.5);
       sp:=sin(eulerZYX[1]*0.5);
       output^.w:=cy*cr*cp+sy*sr*sp;
       output^.v[0]:=cy*sr*cp-sy*cr*sp;
       output^.v[1]:=cy*cr*sp+sy*sr*cp;
       output^.v[2]:=sy*cr*cp-cy*sr*sp;
     end;
end;

function copysign(x:cDouble;y:cDouble):cDouble;
begin
  If ((x>=0) And (y<0)) Or ((x<0) And (y>=0)) Then Result:=-1*x Else Result:=x;
end;


procedure Quaternion_toEulerZYX(q: PQuaternion; output: TDouble3);
Var
  sinr_cosp:cDouble;
  cosr_cosp:cDouble;
  sinp:cDouble;
  siny_cosp:cDouble;
  cosy_cosp:cDouble;
begin
  If Assigned(@output[0]) Then
     Begin
       sinr_cosp:=+2.0*(q^.w*q^.v[0]+q^.v[1]*q^.v[2]);
       cosr_cosp:=+1.0-2.0*(sqr(q^.v[0])+sqr(q^.v[1]));
       output[0]:=arctan2(sinr_cosp,cosr_cosp);
       sinp:=+2.0*(q^.w*q^.v[1]-q^.v[2]*q^.v[0]);
       If abs(sinp)>=1 Then
         output[1]:=copysign(Pi/2,sinp)
       Else
         output[1]:=arcsin(sinp);
       siny_cosp:=+2.0*(q^.w*q^.v[2]+q^.v[0]*q^.v[1]);
       cosy_cosp:=+1.0-2.0*(sqr(q^.v[1])+sqr(q^.v[2]));
       output[2]:=arctan2(siny_cosp,cosy_cosp);
     end;
end;

procedure Quaternion_fromXRotation(angle: cDouble; output: PQuaternion);
Var
  axis:Array[0..2] Of cDouble = (1.0,0,0);
begin
  If Assigned(output) Then Quaternion_fromAxisAngle(axis,angle,output);
end;

procedure Quaternion_fromYRotation(angle: cDouble; output: PQuaternion);
Var
  axis:Array[0..2] Of cDouble = (0,1.0,0);
begin
  If Assigned(output) Then Quaternion_fromAxisAngle(axis,angle,output);
end;

procedure Quaternion_fromZRotation(angle: cDouble; output: PQuaternion);
Var
  axis:Array[0..2] Of cDouble = (0,0,1.0);
begin
  If Assigned(output) Then Quaternion_fromAxisAngle(axis,angle,output);
end;

function Quaternion_norm(q: PQuaternion): cDouble;
begin
  If Assigned(q) Then Result:=sqrt(sqr(q^.w)+sqr(q^.v[0])+sqr(q^.v[1])+sqr(q^.v[2]));
end;

procedure Quaternion_normalize(q: PQuaternion; output: PQuaternion);
Var
  len:cDouble;
begin
  If Assigned(output) Then
     Begin
       len:=Quaternion_norm(q);
       Quaternion_set(q^.w/len,q^.v[0]/len,q^.v[1]/len,q^.v[2]/len,output);
     end;
end;

procedure Quaternion_conjugate(q: PQuaternion; output: PQuaternion);
begin
  If Assigned(output) Then
     Begin
       output^.w:=q^.w;
       output^.v[0]:=-q^.v[0];
       output^.v[1]:=-q^.v[1];
       output^.v[2]:=-q^.v[2];
     end;
end;

procedure Quaternion_multiply(q1: PQuaternion; q2: PQuaternion;
  output: PQuaternion);
Var
  res:TQuaternion; //Local var neccessary?
begin
  If Assigned(output) Then
     Begin
       res.w:=   q1^.w   *q2^.w    - q1^.v[0]*q2^.v[0] - q1^.v[1]*q2^.v[1] - q1^.v[2]*q2^.v[2];
       res.v[0]:=q1^.v[0]*q2^.w    + q1^.w   *q2^.v[0] + q1^.v[1]*q2^.v[2] - q1^.v[2]*q2^.v[1];
       res.v[1]:=q1^.w   *q2^.v[1] - q1^.v[0]*q2^.v[2] + q1^.v[1]*q2^.w    + q1^.v[2]*q2^.v[0];
       res.v[2]:=q1^.w   *q2^.v[2] + q1^.v[0]*q2^.v[1] - q1^.v[1]*q2^.v[0] + q1^.v[2]*q2^.w;
       output^:=res;
     end;
end;

procedure Quaternion_rotate(q: PQuaternion; v: TDouble3; output: TDouble3);
Var
  res:TDouble3;  //Local var neccessary?
  ww:cDouble;
  xx:cDouble;
  yy:cDouble;
  zz:cDouble;
  wx:cDouble;
  wy:cDouble;
  wz:cDouble;
  xy:cDouble;
  xz:cDouble;
  yz:cDouble;
begin
  If Assigned(@output[0]) Then
     Begin
       ww:=sqr(q^.w);
       xx:=sqr(q^.v[0]);
       yy:=sqr(q^.v[1]);
       zz:=sqr(q^.v[2]);
       wx:=q^.w*q^.v[0];
       wy:=q^.w*q^.v[1];
       wz:=q^.w*q^.v[2];
       xy:=q^.v[0]*q^.v[1];
       xz:=q^.v[0]*q^.v[2];
       yz:=q^.v[1]*q^.v[2];
       res[0]:=ww*v[0]+2*wy*v[2]-2*wz*v[1]+xx*v[0]+2*xy*v[1]+2*xz*v[2]-zz*v[0]-yy*v[0];
       res[1]:=2*xy*v[0]+yy*v[1]+2*yz*v[2]+2*wz*v[0]-zz*v[1]+ww*v[1]-2*wx*v[2]-xx*v[1];
       res[2]:=2*xz*v[0]+2*yz*v[1]+zz*v[2]-2*wy*v[0]-yy*v[2]+2*wx*v[1]-xx*v[2]+ww*v[2];
       output[0]:=res[0];
       output[1]:=res[1];
       output[2]:=res[2];
     end;
end;

procedure Quaternion_slerp(q1: PQuaternion; q2: PQuaternion; t: cDouble;
  output: PQuaternion);
Var
  res:TQuaternion; //Local var neccessary?
  cosHalfTheta:cDouble;
  halfTheta:cDouble;
  sinHalfTheta:cDouble;
  ratioA:cDouble;
  ratioB:cDouble;
begin
  If Assigned(output) Then   //In original c-code there is no assert if output!=Null  ??!?!?!?
    Begin
      cosHalfTheta:=q1^.w*q2^.w+q1^.v[0]*q2^.v[0]+q1^.v[1]*q2^.v[1]+q1^.v[2]*q2^.v[2];
      If abs(cosHalfTheta)>=1.0 Then
        begin
          Quaternion_copy(q1,output);
          Exit;
        end;
      halfTheta:=arccos(cosHalfTheta);
      sinHalfTheta:=sqrt(1.0-sqr(cosHalfTheta));
      If abs(sinHalfTheta)<QUATERNION_EPS Then
        Begin
          res.w:=0.5*(q1^.w+q2^.w);
          res.v[0]:=0.5*(q1^.v[0]+q2^.v[0]);
          res.v[1]:=0.5*(q1^.v[1]+q2^.v[1]);
          res.v[2]:=0.5*(q1^.v[2]+q2^.v[2]);
        end;
      ratioA:=sin((1-t)*halfTheta)/sinHalfTheta;
      ratioB:=sin(t*halfTheta)/sinHalfTheta;
      res.w:=q1^.w*ratioA+q2^.w*ratioB;
      res.v[0]:=q1^.v[0]*ratioA+q2^.v[0]*ratioB;
      res.v[1]:=q1^.v[1]*ratioA+q2^.v[1]*ratioB;
      res.v[2]:=q1^.v[2]*ratioA+q2^.v[2]*ratioB;
      output^:=res;
    end;
end;

end.

