program testquaternion;
{$mode objfpc}{$H+}
//{$DEFINE UsecTypes}

uses sysutils,math,quaternion{$IFDEF UsecTypes},ctypes{$ENDIF};
{$IFNDEF UsecTypes}
Type
  cDouble = Double;
  cBool = Boolean;
{$ENDIF}

Procedure ASSERT_SAME_DOUBLE(description:PChar;x:cDouble;y:cDouble);
Begin
  If abs(x-y)>QUATERNION_EPS Then Writeln('TEST FAILED: ',Format('%s (%f != %f)',[description,x,y]));
End;

Procedure ASSERT_TRUE(description:PChar;check:cBool);
Begin
  If Not check Then Writeln('TEST FAILED: ',Format('%s',[description]));
End;

Procedure ASSERT_FALSE(description:PChar;check:cBool);
Begin
  ASSERT_TRUE(description,Not check);
End;

Procedure testQuaternion_set;
Var
  q:TQuaternion;
Begin
  Quaternion_set(5.1,4.2,3.3,2.4,@q);
  ASSERT_SAME_DOUBLE('Quaternion_set should set w', q.w, 5.1);
  ASSERT_SAME_DOUBLE('Quaternion_set should set v[0]', q.v[0], 4.2);
  ASSERT_SAME_DOUBLE('Quaternion_set should set v[1]', q.v[1], 3.3);
  ASSERT_SAME_DOUBLE('Quaternion_set should set v[2]', q.v[2], 2.4);
End;

Procedure testQuaternion_setIdentity;
Var
  q:TQuaternion;
Begin
  Quaternion_set(5.1,4.2,3.3,2.4,@q);
  Quaternion_setIdentity(@q);
  ASSERT_SAME_DOUBLE('Quaternion_setIdentity has wrong w', q.w, 1);
  ASSERT_SAME_DOUBLE('Quaternion_setIdentity has wrong v[0]', q.v[0], 0);
  ASSERT_SAME_DOUBLE('Quaternion_setIdentity has wrong v[1]', q.v[1], 0);
  ASSERT_SAME_DOUBLE('Quaternion_setIdentity has wrong v[2]', q.v[2], 0);
End;

Procedure testQuaternion_copy;
Var
  q:TQuaternion;
  r:TQuaternion;
Begin
  Quaternion_set(5.1,4.2,3.3,2.4,@q);
  Quaternion_copy(@q, @r);
  ASSERT_SAME_DOUBLE('Quaternion_copy has wrong w', r.w, q.w);
  ASSERT_SAME_DOUBLE('Quaternion_copy has wrong v[0]', r.v[0], q.v[0]);
  ASSERT_SAME_DOUBLE('Quaternion_copy has wrong v[1]', r.v[1], q.v[1]);
  ASSERT_SAME_DOUBLE('Quaternion_copy has wrong v[2]', r.v[2], q.v[2]);
End;

Procedure testQuaternion_conjugate;
Var
  q:TQuaternion;
  c:TQuaternion;
Begin
  Quaternion_set(5.1,4.2,3.3,2.4,@q);
  Quaternion_conjugate(@q, @c);
  ASSERT_SAME_DOUBLE('Quaternion_conjugate should keep w', c.w, q.w);
  ASSERT_SAME_DOUBLE('Quaternion_conjugate should negate v[0]',c.v[0], -q.v[0]);
  ASSERT_SAME_DOUBLE('Quaternion_conjugate should negate v[1]',c.v[1], -q.v[1]);
  ASSERT_SAME_DOUBLE('Quaternion_conjugate should negate v[2]',c.v[2], -q.v[2]);
End;

Procedure testQuaternion_equal;
Var
  q1:TQuaternion;
  q2:TQuaternion;
  q3:TQuaternion;
  q4:TQuaternion;
  q5:TQuaternion;
Begin
  Quaternion_set(5.1, 4.2, 3.3, 2.4, @q1);
  Quaternion_set(5.1, 4.2, 3.3, 2.4, @q2);
  Quaternion_set(5.1, 4.2 + (QUATERNION_EPS/2.0), 3.3, 2.4, @q3);
  Quaternion_set(9.1, 4.2, 3.3, 2.4, @q4);
  Quaternion_set(5.1, 4.2, 9.3, 2.4, @q5);
  ASSERT_TRUE('Quaternion_equal with same reference', Quaternion_equal(@q1, @q1));
  ASSERT_TRUE('Quaternion_equal with same values', Quaternion_equal(@q1, @q2));
  ASSERT_TRUE('Quaternion_equal with similar values', Quaternion_equal(@q1, @q3));
  ASSERT_FALSE('Quaternion_equal with different w', Quaternion_equal(@q1, @q4));
  ASSERT_FALSE('Quaternion_equal with different v', Quaternion_equal(@q1, @q5));
End;

Procedure testQuaternion_norm;
Var
  q:TQuaternion;
  c:TQuaternion;
Begin
  Quaternion_set(0.7071, 0.7071, 0, 0, @q);
  Quaternion_setIdentity(@c);
  ASSERT_SAME_DOUBLE('Quaternion_norm of valid q == 1', Quaternion_norm(@q), 1);
  ASSERT_SAME_DOUBLE('Quaternion_norm of identity', Quaternion_norm(@c), 1);
  Quaternion_set(4, 5, 6, 7, @q);
  ASSERT_SAME_DOUBLE('Quaternion_norm larger 1', Quaternion_norm(@q), sqrt(126));
End;

Procedure testQuaternion_normalize;
Var
  q:TQuaternion;
  c:TQuaternion;
Begin
  Quaternion_set(0.7071, 0.7071, 0, 0, @q);
  Quaternion_normalize(@q, @c);
  ASSERT_SAME_DOUBLE('Quaternion_normalize should stay 1', Quaternion_norm(@c), 1);
  Quaternion_set(4, 5, 6, 7, @q);
  Quaternion_normalize(@q, @c);
  ASSERT_SAME_DOUBLE('Quaternion_normalize was not 1', Quaternion_norm(@q), sqrt(126));
  ASSERT_SAME_DOUBLE('Quaternion_normalize is 1 afterwards', Quaternion_norm(@c), 1);
End;

Procedure testQuaternion_fromAxisAngle;
Var
  q:TQuaternion;
  vector:Array[0..2] Of cDouble=(1,0,0);
  angle:cDouble;
Begin
  angle:=degtorad(90.0);
  Quaternion_fromAxisAngle(vector, angle, @q);
  ASSERT_SAME_DOUBLE('Quaternion_fromAxisAngle has wrong w', q.w, 0.7071);
  ASSERT_SAME_DOUBLE('Quaternion_fromAxisAngle has wrong v[0]', q.v[0], 0.7071);
  ASSERT_SAME_DOUBLE('Quaternion_fromAxisAngle has wrong v[1]', q.v[1], 0);
  ASSERT_SAME_DOUBLE('Quaternion_fromAxisAngle has wrong v[2]', q.v[2], 0);
End;

Procedure testQuaternion_fromXRotation;
Var
  q:TQuaternion;
  vector:Array[0..2] Of cDouble=(1,0,0);
  angle:cDouble;
Begin
  angle:=degtorad(90.0);
  Quaternion_fromAxisAngle(vector, angle, @q);
  ASSERT_SAME_DOUBLE('Quaternion_fromXRotation  has wrong w', q.w, 0.7071);
  ASSERT_SAME_DOUBLE('Quaternion_fromXRotation  has wrong v[0]', q.v[0], 0.7071);
  ASSERT_SAME_DOUBLE('Quaternion_fromXRotation  has wrong v[1]', q.v[1], 0);
  ASSERT_SAME_DOUBLE('Quaternion_fromXRotation  has wrong v[2]', q.v[2], 0);
end;

Procedure testQuaternion_fromYRotation;
Var
  q:TQuaternion;
  vector:Array[0..2] Of cDouble=(0,1,0);
  angle:cDouble;
Begin
  angle:=degtorad(90.0);
  Quaternion_fromAxisAngle(vector, angle, @q);
  ASSERT_SAME_DOUBLE('Quaternion_fromYRotation  has wrong w', q.w, 0.7071);
  ASSERT_SAME_DOUBLE('Quaternion_fromYRotation  has wrong v[0]', q.v[0], 0);
  ASSERT_SAME_DOUBLE('Quaternion_fromYRotation  has wrong v[1]', q.v[1], 0.7071);
  ASSERT_SAME_DOUBLE('Quaternion_fromYRotation  has wrong v[2]', q.v[2], 0);
end;

Procedure testQuaternion_fromZRotation;
Var
  q:TQuaternion;
  vector:Array[0..2] Of cDouble=(0,0,1);
  angle:cDouble;
Begin
  angle:=degtorad(90.0);
  Quaternion_fromAxisAngle(vector, angle, @q);
  ASSERT_SAME_DOUBLE('Quaternion_fromZRotation  has wrong w', q.w, 0.7071);
  ASSERT_SAME_DOUBLE('Quaternion_fromZRotation  has wrong v[0]', q.v[0], 0);
  ASSERT_SAME_DOUBLE('Quaternion_fromZRotation  has wrong v[1]', q.v[1], 0);
  ASSERT_SAME_DOUBLE('Quaternion_fromZRotation  has wrong v[2]', q.v[2], 0.7071);
end;

Procedure testQuaternion_toAxisAngle;
Var
  v90:Array[0..2] Of cDouble;
  rot90:TQuaternion;
  a90:cDouble;
  v0:Array[0..2] Of cDouble;
  rot0:TQuaternion;
  a0:cDouble;
  arbAxis:Array[0..2] Of cDouble=(0.802, 0.267, -0.534);
  vArb:Array[0..2] Of cDouble;
  rotArb:TQuaternion;
  aArb:cDouble;
Begin
  Quaternion_set(0.7071, 0.7071, 0, 0, @rot90);
  a90:=Quaternion_toAxisAngle(@rot90, v90);
  ASSERT_SAME_DOUBLE('Quaternion_toAxisAngle should calculate 90 deg angle', a90, degtorad(90.0));
  ASSERT_SAME_DOUBLE('Quaternion_toAxisAngle should calculate 90 deg X axis', v90[0], 1);
  ASSERT_SAME_DOUBLE('Quaternion_toAxisAngle should calculate 90 deg Y axis', v90[1], 0);
  ASSERT_SAME_DOUBLE('Quaternion_toAxisAngle should calculate 90 deg Z axis', v90[2], 0);
  Quaternion_setIdentity(@rot0);
  a0:=Quaternion_toAxisAngle(@rot0, v0);
  ASSERT_SAME_DOUBLE('Quaternion_toAxisAngle should calculate 0 deg angle', a0, 0);
  Quaternion_fromAxisAngle(arbAxis, 1.11, @rotArb);
  aArb:=Quaternion_toAxisAngle(@rotArb, vArb);
  ASSERT_SAME_DOUBLE('Quaternion_toAxisAngle should inverse fromAxisAngle() (angle)', aArb, 1.11);
  ASSERT_SAME_DOUBLE('Quaternion_toAxisAngle should inverse fromAxisAngle() (X axis)', vArb[0], arbAxis[0]);
  ASSERT_SAME_DOUBLE('Quaternion_toAxisAngle should inverse fromAxisAngle() (Y axis)', vArb[1], arbAxis[1]);
  ASSERT_SAME_DOUBLE('Quaternion_toAxisAngle should inverse fromAxisAngle() (Z axis)', vArb[2], arbAxis[2]);
End;

Procedure testQuaternion_fromEulerZYX;
Var
  e0:Array[0..2] Of cDouble;
  e1:Array[0..2] Of cDouble;
  e2:Array[0..2] Of cDouble;
  e3:Array[0..2] Of cDouble;
  q:TQuaternion;
  r:TQuaternion;
Begin
  e0:=Default(TDouble3);
  e0[0]:=degtorad(90.0);
  Quaternion_set(0.7071, 0.7071, 0, 0, @r);
  Quaternion_fromEulerZYX(e0, @q);
  ASSERT_TRUE('Quaternion_fromEulerZYX for x-rotation only', Quaternion_equal(@q, @r));
  e1:=Default(TDouble3);
  e1[0]:=degtorad(90.0);
  e1[2]:=degtorad(180.0);
  Quaternion_set(0, 0, 0.7071, 0.7071,@r);
  Quaternion_fromEulerZYX(e1, @q);
  ASSERT_TRUE('Quaternion_fromEulerZYX example 1', Quaternion_equal(@q,@r));
  e2:=Default(TDouble3);
  e2[0]:=degtorad(90.0);
  e2[2]:=degtorad(90.0);
  Quaternion_set(0.5, 0.5, 0.5, 0.5, @r);
  Quaternion_fromEulerZYX(e2, @q);
  ASSERT_TRUE('Quaternion_fromEulerZYX example 2', Quaternion_equal(@q,@r));
  e3[0]:=degtorad(165.0);
  e3[1]:=degtorad(63.0);
  e3[2]:=degtorad(122.0);
  Quaternion_set(0.5070333, 0.3501829, 0.7724199, -0.1538071, @r);
  Quaternion_fromEulerZYX(e3, @q);
  ASSERT_TRUE('Quaternion_fromEulerZYX example 3', Quaternion_equal(@q, @r));
End;

Procedure testQuaternion_toEulerZYX;
Var
  euler:Array[0..2] Of cDouble;
  q:TQuaternion;
Begin
  Quaternion_set(0.7071, 0.7071, 0, 0, @q);
  Quaternion_toEulerZYX(@q, euler);
  ASSERT_SAME_DOUBLE('Quaternion_toEulerZYX x-rotation only (X axis)', euler[0], degtorad(90.0));
  ASSERT_SAME_DOUBLE('Quaternion_toEulerZYX x-rotation only (Y axis)', euler[1], 0);
  ASSERT_SAME_DOUBLE('Quaternion_toEulerZYX x-rotation only (Z axis)', euler[2], 0);
  Quaternion_set(0, 0, 0.7071, 0.7071, @q);
  Quaternion_toEulerZYX(@q, euler);
  ASSERT_SAME_DOUBLE('Quaternion_toEulerZYX xz-rotation only (X axis)', euler[0], degtorad(90.0));
  ASSERT_SAME_DOUBLE('Quaternion_toEulerZYX xz-rotation only (Y axis)', euler[1], 0);
  ASSERT_SAME_DOUBLE('Quaternion_toEulerZYX xz-rotation only (Z axis)', euler[2], degtorad(180.0));
  Quaternion_set(0.5, 0.5, 0.5, 0.5, @q);
  Quaternion_toEulerZYX(@q, euler);
  ASSERT_SAME_DOUBLE('Quaternion_toEulerZYX xy-rotation (X axis)', euler[0], degtorad(90.0));
  ASSERT_SAME_DOUBLE('Quaternion_toEulerZYX xy-rotation (Y axis)', euler[1], 0);
  ASSERT_SAME_DOUBLE('Quaternion_toEulerZYX xy-rotation (Z axis)', euler[2], degtorad(90.0));
  Quaternion_set(0.5070333, 0.3501829, 0.7724199, -0.1538071, @q);
  Quaternion_toEulerZYX(@q, euler);
  ASSERT_SAME_DOUBLE('Quaternion_toEulerZYX xy-rotation (X axis)', euler[0], degtorad(165.0));
  ASSERT_SAME_DOUBLE('Quaternion_toEulerZYX xy-rotation (Y axis)', euler[1], degtorad(63.0));
  ASSERT_SAME_DOUBLE('Quaternion_toEulerZYX xy-rotation (Z axis)', euler[2], degtorad(122.0));
End;

Procedure testQuaternion_multiply;
Var
  q1:TQuaternion;
  q2:TQuaternion;
  res:TQuaternion;
  r:TQuaternion;
Begin
  Quaternion_set(1.0, 0.0,  1.0, 0.0,  @q1);
  Quaternion_set(1.0, 0.5,  0.5, 0.75, @q2);
  Quaternion_set(0.5, 1.25, 1.5, 0.25, @r);
  Quaternion_multiply(@q1, @q2, @res);
  ASSERT_TRUE('Quaternion_multiply example 1', Quaternion_equal(@res, @r));
  Quaternion_set(1.0, 0.0,  1.0, 0.0,  @q1);
  Quaternion_set(0, 0, 2, 0, @r);
  Quaternion_multiply(@q1, @q1, @res);
  ASSERT_TRUE('Quaternion_multiply example 2', Quaternion_equal(@res, @r));
  Quaternion_set(1.0, 0.0,  1.0, 0.0,  @q1);
  Quaternion_set(2.0, 1.0, 0.1,  0.1, @q2);
  Quaternion_set(1.9, 1.1, 2.1, -0.9, @r);
  Quaternion_multiply(@q1, @q2, @res);
  ASSERT_TRUE('Quaternion_multiply example 3', Quaternion_equal(@res, @r));
End;

Procedure testQuaternion_rotate;
Var
  res:Array[0..2] Of cDouble;
  v1:Array[0..2] Of cDouble=(5.1, 6.8, -5.3);
  identity:TQuaternion;
  q:TQuaternion;
  v2:Array[0..2] Of cDouble=(1, 0, 0);
  v3:Array[0..2] Of cDouble=(1, 0, 0);
  v4:Array[0..2] Of cDouble=(1, 0, 0);
Begin
  Quaternion_setIdentity(@identity);
  Quaternion_rotate(@identity, v1, res);
  ASSERT_SAME_DOUBLE('Quaternion_rotate example 1 (X-axis)', res[0], v1[0]);
  ASSERT_SAME_DOUBLE('Quaternion_rotate example 1 (Y-axis)', res[1], v1[1]);
  ASSERT_SAME_DOUBLE('Quaternion_rotate example 1 (Z-axis)', res[2], v1[2]);
  Quaternion_set(0.5, 0.5, 0.5, 0.5, @q);
  Quaternion_rotate(@q, v2, res);
  ASSERT_SAME_DOUBLE('Quaternion_rotate example 2 (X-axis)', res[0], 0);
  ASSERT_SAME_DOUBLE('Quaternion_rotate example 2 (Y-axis)', res[1], 1);
  ASSERT_SAME_DOUBLE('Quaternion_rotate example 2 (Z-axis)', res[2], 0);
  Quaternion_set(0.6532815, -0.270598, 0.270598, 0.6532815, @q);
  Quaternion_rotate(@q, v3, res);
  ASSERT_SAME_DOUBLE('Quaternion_rotate example 3 (X-axis)', res[0], 0);
  ASSERT_SAME_DOUBLE('Quaternion_rotate example 3 (Y-axis)', res[1], +0.7071);
  ASSERT_SAME_DOUBLE('Quaternion_rotate example 3 (Z-axis)', res[2], -0.7071);
  Quaternion_set(0.7071, 0, 0, 0.7071, @q);
  Quaternion_rotate(@q, v4, res);
  ASSERT_SAME_DOUBLE('Quaternion_rotate example 4 (X-axis)', res[0], 0);
  ASSERT_SAME_DOUBLE('Quaternion_rotate example 4 (Y-axis)', res[1], 1);
  ASSERT_SAME_DOUBLE('Quaternion_rotate example 4 (Z-axis)', res[2], 0);
End;

Procedure testQuaternion_slerp;
Var
  q1:TQuaternion;
  q2:TQuaternion;
  res:TQuaternion;
Begin
  Quaternion_set(0.6532815, -0.270598, 0.270598, 0.6532815, @q1);
  Quaternion_set(0.5, 0.5, 0.5, 0.5, @q2);
  Quaternion_slerp(@q1, @q2, 0, @res);
  ASSERT_TRUE('Quaternion_slerp with t=0', Quaternion_equal(@res, @q1));
  Quaternion_slerp(@q1, @q2, 1, @res);
  ASSERT_TRUE('Quaternion_slerp with t=1', Quaternion_equal(@res, @q2));
  Quaternion_slerp(@q1, @q2, 0.62, @res);
  ASSERT_SAME_DOUBLE('Quaternion_slerp with t=0.62 (w)', res.w, 0.6119266025696755);
  ASSERT_SAME_DOUBLE('Quaternion_slerp with t=0.62 (v[0])', res.v[0], 0.22069444274723088);
  ASSERT_SAME_DOUBLE('Quaternion_slerp with t=0.62 (v[1])', res.v[1], 0.4498729015909088);
  ASSERT_SAME_DOUBLE('Quaternion_slerp with t=0.62 (v[2])', res.v[2], 0.6119266025696755);
End;

begin
  testQuaternion_set;
  testQuaternion_setIdentity;
  testQuaternion_copy;
  testQuaternion_equal;
  testQuaternion_conjugate;
  testQuaternion_norm;
  testQuaternion_normalize;
  testQuaternion_fromAxisAngle;
  testQuaternion_fromXRotation;
  testQuaternion_fromYRotation;
  testQuaternion_fromZRotation;
  testQuaternion_toAxisAngle;
  testQuaternion_fromEulerZYX;
  testQuaternion_toEulerZYX;
  testQuaternion_multiply;
  testQuaternion_rotate;
  testQuaternion_slerp;
  Readln;
end.

